/* *****************************************************************************
 * type/caspro.c -- get the newer CAS type out of raw identification data.
 * Copyright (C) 2017 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
 *
 * This file is part of libg1m.
 * libg1m is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License,
 * or (at your option) any later version.
 *
 * libg1m is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with libg1m; if not, see <http://www.gnu.org/licenses/>.
 * ************************************************************************** */
#include <libg1m/internals.h>

/* ************************************************************************** */
/*  Local types                                                               */
/* ************************************************************************** */
/* Correspondance type */
#define flg_type  0x0001 /* if not there, the type is defined by the datatype */
#define flg_flags  0x0002 /* the following value is the flag to OR. */
struct app_corresp {
	/* identification */
	const char *name;

	/* values */
	unsigned int flags; /* the internal flags */
	unsigned long int value; /* meaning of this depends on the flags */
};

/* ************************************************************************** */
/*  Correspondances                                                           */
/* ************************************************************************** */
/* All correspondances. Remarks:
 * - Correspondances with a NULL data type means the data type isn't to be
 *   read. */

#define TTERM {NULL, 0, 0}
static struct app_corresp apps[] = {
	{"TXT", 0,         0}, /* editor */
	{"VAL", 0,         0}, /* RUN? */
	{"IMG", 0,         0}, /* picture */
	{"REQ", flg_flags, g1m_mcsflag_request}, /* request */
	{"END", flg_type,  g1m_mcstype_end}, /* end */

	/* terminating entry */
	TTERM
};

/* ************************************************************************** */
/*  Main functions                                                            */
/* ************************************************************************** */
/**
 *	g1m_maketype_caspro:
 *	Get the libg1m MCS type from raw CAS identification data.
 *
 *	@arg	head		the head to fill.
 *	@arg	app			the app string.
 *	@arg	datatype	the data type string.
 *	@return				the error (if any).
 */

int g1m_maketype_caspro(g1m_mcshead_t *head,
	const char *app, const char *datatype)
{
	/* look for the app */
	struct app_corresp *a;
	for (a = apps; a->name; a++) {
		if (!strncmp(a->name, app, 3))
			break;
	}
	if (!a->name) goto notfound;

	/* check if we need the datatype */
	if (a->flags & flg_type) {
		memset(head, 0, sizeof(g1m_mcshead_t));
		head->type = a->value;
		memcpy(head->_appname, app, 3); head->_appname[3] = 0;
	} else {
		int err = g1m_maketype_cas(head, datatype);
		if (err) return (err);
	}

	/* copy raw information */
	head->flags = (head->flags & ~g1m_mcsmask_info) | g1m_mcsinfo_caspro;
	memcpy(head->_appname, app, 3); head->_appname[3] = 0;
	if (a->flags & flg_flags)
		head->flags |= (unsigned int)a->value;

	/* fill in info and return */
	return (0);

notfound:
	log_info("Type with '%.3s' app name and '%.2s' data string was not"
		"implemented or recognized.", app, datatype);
	head->type = 0;
	return (1);
}
